import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;

public class RuleTable extends JPanel implements MouseListener, MouseMotionListener
{
  private final int border = 4;
  
  private int rows;
  private int columns;
  private int squareSize;
  private int grid[][];
  private BufferedImage image;
  private boolean canChange = true;
  private SquareCell squareCell;
  private ColorSelector colorSelector;
  private StatusBar statusBar;

  public RuleTable(int rows, int columns, SquareCell squareCell, boolean canChange)
  {
    this(rows, columns, squareCell);
    this.canChange = canChange;
  }

  public RuleTable(int rows, int columns, SquareCell squareCell)
  {
    this.rows = rows;
    this.columns = columns;
    squareSize = 20;
    this.squareCell = squareCell;
    this.colorSelector = squareCell.getColorSelector();
    this.statusBar = squareCell.getStatusBar();
    grid = new int[rows][columns];
    setOpaque(true);
    setDoubleBuffered(false);
    setPreferredSize(new Dimension(columns * squareSize + 2 * border, rows * squareSize + 2 * border));
    addMouseListener(this);
    addMouseMotionListener(this);
  }

  public int getRows()
  {
    return rows;
  }

  public int getColumns()
  {
    return columns;
  }  

  public int xCoordToColumnNumber(int x)
  {
    int c = (x - border) / squareSize;
    if (c < 0)
      return 0;
    else if (c >= columns)
      return columns - 1;
    else
      return c;
  }
 
  public int yCoordToRowNumber(int y)
  {
    int r = (y - border) / squareSize;
    if (r < 0)
      return 0;
    else if (r >= rows)
      return rows - 1;
    else
      return r;
  }

  public void setState(int r, int c, int state)
  {
    grid[r][c] = state;
  }

  public int getState(int r, int c)
  {
    return grid[r][c];
  }

  public void clear()
  {
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        grid[r][c] = 0;
  }

  public int[][] getGrid()
  {
    return grid;
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    if ((image == null) || image.getWidth() != getWidth() || image.getHeight() != getHeight())
      image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
    drawSquares();
    g.drawImage(image, 0, 0, null);
  }

  private void drawSquares()
  {
    Graphics g = image.getGraphics();
    g.setColor(Color.gray);
    g.fillRect(0, 0, image.getWidth(), image.getHeight());
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        drawSquare(g, r, c, colorSelector.getColor(grid[r][c]));
    g.dispose();
  }

  private void drawSquare(Graphics g, int r, int c, Color color)
  {
    int y = border + (r * squareSize);
    int h = squareSize;
    int x = border + (c * squareSize);
    int w = squareSize;
    g.setColor(colorSelector.getGridColor());
    g.drawRect(x, y, squareSize, squareSize);
    g.setColor(color);
    g.fillRect(x + 1, y + 1, squareSize - 1, squareSize - 1);
  }

  public void mousePressed(MouseEvent e)
  {
    if (! canChange)
      return;
    int r = yCoordToRowNumber(e.getY());
    int c = xCoordToColumnNumber(e.getX());
    int s = squareCell.getCurrentState();
    if (e.isShiftDown())
    {
      s = (s + 1) % squareCell.getNoStates();
      squareCell.setCurrentState(s);
      colorSelector.showCurrentState();
    }
    statusBar.setCurrentState(s);
    setState(r, c, s);
    repaint();
  }

  public void mouseMoved(MouseEvent e)
  {
    int r = yCoordToRowNumber(e.getY());
    int c = xCoordToColumnNumber(e.getX());
    statusBar.setCursorLocation(r, c);
    statusBar.setCurrentState(getState(r, c));
  }

  public void mouseExited(MouseEvent e)
  {
    statusBar.removeCursorLocation();
    statusBar.removeCurrentState();
  }

  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}
  public void mouseDragged(MouseEvent e) {}
}
