import java.awt.*;
import java.io.*;
import java.util.*;

public class Selector
{
  private SquareCell squareCell;

  public enum State {None, TopLeftMarked, GroupSelected};

  private byte selection[][];
  private State state;
  private int topLeftRow;
  private int topLeftColumn;
  private int bottomRightRow;
  private int bottomRightColumn;
  private int rows;
  private int columns;
  private boolean cut = true;

  public Selector(SquareCell squareCell)
  {
    this.squareCell = squareCell;
    state = State.None;
  }

  public void setState(State state)
  {
    this.state = state;
  }

  public State getState()
  {
    return state;
  }

  public void setCut(boolean cut)
  {
    this.cut = cut;
  }

  public boolean getCut()
  {
    return cut;
  }

  public int getCellState(int r, int c)
  {
    return selection[r][c];
  }

  public void setTopLeft(int r, int c)
  {
    topLeftRow = r;
    topLeftColumn = c;
    bottomRightRow = r;
    bottomRightColumn = c;
  }

  public void makeNewSelector(int row, int column, int rowOffset, int columnOffset)
  {
    int t;
    int r = topLeftRow;
    int c = topLeftColumn;
    if (r > row)
    {
      t = r;
      r = row;
      row = t;
    }
    if (c > column)
    {
      t = c;
      c = column;
      column = t;
    }
    topLeftRow = r;
    topLeftColumn = c;
    bottomRightRow = row;
    bottomRightColumn = column;
    rows = bottomRightRow - topLeftRow + 1;
    columns = bottomRightColumn - topLeftColumn + 1;
    selection = new byte[rows][columns];
  }

  public void save(int row, int column, int rowOffset, int columnOffset, byte cells[][])
  {
    makeNewSelector(row, column, rowOffset, columnOffset);
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        selection[r][c] = cells[rowOffset + topLeftRow + r][columnOffset + topLeftColumn + c];
    if (cut)
    {
      for (int r = topLeftRow; r <= bottomRightRow; r++)
        for (int c = topLeftColumn; c <= bottomRightColumn; c++)
          cells[rowOffset + r][columnOffset + c] = 0;
    }
  }

  public void restore(int cellRows, int cellColumns, int rowOffset, int columnOffset, byte cells[][])
  {
    int state;
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
      { 
        state = selection[r][c];
        int cellRow = rowOffset + topLeftRow + r;
        int cellColumn = columnOffset + topLeftColumn + c;
        if ((state > 0) && (cellRow >= 0) && (cellRow < cellRows) && (cellColumn >= 0) && (cellColumn < cellColumns))
          cells[cellRow][cellColumn] = (byte)state;
      }
    selection = null;
  }

  public void resetSelector()
  {
    state = State.None;
    selection = null;
  }

  public void readSelection(BufferedReader br)
  {
    int r, c;
    int n;
    char ch;
    String s = "";
    try
    {
      String line = br.readLine();
      while (line != null)
        if (line.startsWith("#L "))
        {
          s = s + line.substring(3);
          line = br.readLine();
        }
        else
          return;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    r = 0;
    c = 0;
    while (s.length() > 0)
    {
      n = 0;
      ch = s.charAt(0);
      while ((ch >= '0') && (ch <= '9'))
      {
        n = n * 10 + (ch - '0');
        s = s.substring(1);
        ch = s.charAt(0);
      }
      if (n == 0)
        n = 1;
      if (ch == '$')
      {
        r += n;
        c = 0;
      }
      else
        if (ch == '.')
          c += n;
        else
        {
          for (int i = 0;  i < n; i++)
          {
            selection[r][c] = (byte)(ch - '@');
            c++;
          }
        }
      s = s.substring(1);
    }
  }

  public void writeSelection(PrintWriter pw)
  {
    int r, c;
    int ce = columns - 1;
    char state;
    char lastState = '0';
    int n = 0;
    boolean blankLine;
    String s = "";
    for (r = 0; r < rows; r++)
    {
      blankLine = true;
      for (c = 0; c < columns; c++)
      {
        if (selection[r][c] > 0)
        {
          blankLine = false;
          break;
        }
      }
      if (blankLine)
      {
        if (lastState == '$')
          n++;
        else
        {
          lastState = '$';
          n = 1;
        }
        continue;
      }
      else
        for (c = columns - 1; c >= 0; c--)
        {
          if (selection[r][c] > 0)
          {
            ce = c;
            break;
          }
        }
      for (c = 0; c <= ce;  c++)
      {
        state = (char)(selection[r][c] + 64);
        if (state == '@')
          state = '.';
        if (n == 0)
        {
          lastState = state;
          n = 1;
        }
        else if (state == lastState)
          n++;
        else
        {
          if (n > 1)
            s = s + n;
          s = s + lastState;
          if (s.length() >= 67)
          {
            pw.println("#L " + s);
            s = "";
          }
          lastState = state;
          n = 1;
        }
      }
      if (lastState == '$')
        n++;
      else
      {
        if (n > 1)
          s = s + n;
        s = s + lastState;
        if (s.length() >= 67)
        {
          pw.println("#L " + s);
          s = "";
        }
        lastState = '$';
        n = 1;
      }
    }
    if (s != "")
      pw.println("#L " + s);
  }

  public int getTopLeftRow()
  {
    return topLeftRow;
  }

  public int getTopLeftColumn()
  {
    return topLeftColumn;
  }

  public int getBottomRightRow()
  {
    return bottomRightRow;
  }

  public int getBottomRightColumn()
  {
    return bottomRightColumn;
  }

  public int getRows()
  {
    return rows;
  }

  public int getColumns()
  {
    return columns;
  }

  public void moveUp(int increment)
  {
    topLeftRow -= increment;
    bottomRightRow -= increment;
  }

  public void moveDown(int increment)
  {
    topLeftRow += increment;
    bottomRightRow += increment;
  }

  public void moveLeft(int increment)
  {
    topLeftColumn -= increment;
    bottomRightColumn -= increment;
  }

  public void moveRight(int increment)
  {
    topLeftColumn += increment;
    bottomRightColumn += increment;
  }

  public void move(int r, int c)
  {
    topLeftRow = r;
    bottomRightRow = rows + topLeftRow - 1;
    topLeftColumn = c;
    bottomRightColumn = columns + topLeftColumn - 1;
  }

  public void clear()
  {
    if (state != State.GroupSelected)
      return;
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        selection[r][c] = 0;
  }

  public void flipX()
  {
    if (state != State.GroupSelected)
      return;
    byte t[][] = new byte[rows][columns];
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        t[rows - r - 1][c] = selection[r][c];
    selection = t;
  }

  public void flipY()
  {
    if (state != State.GroupSelected)
      return;
    byte t[][] = new byte[rows][columns];
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        t[r][columns - c - 1] = selection[r][c];
    selection = t;
  }
  
  public void rotateClockwise()
  {
    if (state != State.GroupSelected)
      return;
    byte t[][] = new byte[columns][rows];
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        t[c][rows - r - 1] = selection[r][c];
    selection = t;
    int i = rows;
    rows = columns;
    columns = i;
    bottomRightRow = rows + topLeftRow - 1;
    bottomRightColumn = columns + topLeftColumn - 1;
  }

  public void rotateAntiClockwise()
  {
    if (state != State.GroupSelected)
      return;
    byte t[][] = new byte[columns][rows];
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        t[columns - c - 1][r] = selection[r][c];
    selection = t;
    int i = rows;
    rows = columns;
    columns = i;
    bottomRightRow = rows + topLeftRow - 1;
    bottomRightColumn = columns + topLeftColumn - 1;
  }
  
  public void random()
  {
    if (state != State.GroupSelected)
      return;
    squareCell.random(rows, columns, selection);
  }
}
