import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class SquareCell extends JPanel implements ActionListener, MouseListener, MouseMotionListener, KeyListener
{
  public static JFrame mainFrame;

  private static int width = 1000;
  private static int height = 800;
  private static int cellSize = 20;
  private static int rows = 1000;
  private static int columns = 1000;

  public static void main(String[] args)
  {
    mainFrame = new JFrame("Square Cell");
    mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    mainFrame.add(new SquareCell());
    mainFrame.pack();
    Toolkit k = Toolkit.getDefaultToolkit();
    Dimension s = k.getScreenSize();
    mainFrame.setLocation((s.width - width) / 2, (s.height - height) / 2);
    mainFrame.setSize(width, height);
    mainFrame.setVisible(true);
  }

  private int rowOffset = 0;
  private int columnOffset = 0;

  private byte cells[][];
  private byte cellsCopy[][];

  private int currentState = 1;
  private Patterns patterns;

  private int cycle = 0;

  private boolean running = false;
  private int delay = 0;

  private boolean analyzing = false;

  private int randomDensity = 50;
  private boolean randomClear = true;
  private boolean randomAll = true;
  private int randomMinState = 0;
  private int randomMaxState = 0;

  private File currentDirectory;

  private Rule rule;

  private SquareGrid display;
  private Selector selector;
  private ColorSelector colorSelector;
  private StatusBar statusBar;
  private NewWorld newWorld;
  private Analyzer analyzer;

  private JToolBar toolBar;

  private JButton quitButton;
  private JButton setNewWorldButton;
  private JButton openFileButton;
  private JButton saveFileButton;

  private JButton newRuleButton;

  private JButton runStopButton;
  private JButton stepButton;
  private JButton clearButton;
  private JButton randomButton;
  private JButton randomOptionsButton;

  private JButton analyzerButton;

  private JButton rulesButton;
  private JButton ruleOptionsButton;

  private JButton changeColorsButton;

  private JButton clearInsideButton;
  private JButton clearOutsideButton;
  private JButton flipXButton;
  private JButton flipYButton;
  private JButton rotateClockwiseButton;
  private JButton rotateAntiClockwiseButton;
  private JButton selectAllButton;
  private JButton centerButton;
  private JButton saveButton;
  private JButton selectionRandomButton;
  private JButton moveCopyButton;

  private Thread runThread;

  public SquareCell()
  {
    UIManager.LookAndFeelInfo[] infos = UIManager.getInstalledLookAndFeels();
    try
    {
      UIManager.setLookAndFeel(infos[0].getClassName());
      SwingUtilities.updateComponentTreeUI(this);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
    cells = new byte[rows][columns];
    cellsCopy = new byte[rows][columns];
    colorSelector = new ColorSelector(mainFrame, this, currentState);
    selector = new Selector(this);
    display = new SquareGrid(rows, columns, cellSize, selector, colorSelector);
    setBorder(BorderFactory.createLineBorder(Color.gray, 3));
    setLayout(new BorderLayout(3, 3));
    add(display, BorderLayout.CENTER);
    makeToolBar();
    add(toolBar, BorderLayout.NORTH);
    statusBar = new StatusBar(rows, columns, display.getRows(), display.getColumns());
    statusBar.setDisplayOffset(rowOffset, columnOffset);
    add(statusBar, BorderLayout.SOUTH);
    display.addMouseListener(this);
    display.addMouseMotionListener(this);
    addKeyListener(this);
    setFocusable(true);
    rule = new Primordia(mainFrame, this);
  }

  private void makeToolBar()
  {
    toolBar = new JToolBar();
    quitButton = getNewButton("Quit", "quit application");
    toolBar.addSeparator();
    setNewWorldButton = getNewButton("NewWorld", "set cellular automata parameters");
    toolBar.addSeparator();
    openFileButton = getNewButton("OpenFile", "read pattern from file to selection");
    saveFileButton = getNewButton("SaveFile", "write pattern to file");
    toolBar.addSeparator();
    newRuleButton = getNewButton("NewRule", "test new random rule");
    toolBar.addSeparator();
    runStopButton = getNewButton("Run", "run the simulation");
    stepButton = getNewButton("Step", "step the simulation");
    toolBar.addSeparator();
    clearButton = getNewButton("Clear", "clear world");
    toolBar.addSeparator();
    randomButton = getNewButton("Random", "set world to random states");
    randomOptionsButton = getNewButton("RandomOptions", "change random options");
    toolBar.addSeparator();
    analyzerButton = getNewButton("Analyzer", "show ship, oscillator or still life characteristics");
    toolBar.addSeparator();
    rulesButton = getNewButton("Rules", "change rule");
    ruleOptionsButton = getNewButton("RuleOptions", "change rule options");
    toolBar.addSeparator();
    changeColorsButton = getNewButton("ChangeColors", "change state colors");
    toolBar.addSeparator();
    clearInsideButton = getNewButton("ClearInside", "clear inside selection");
    clearOutsideButton = getNewButton("ClearOutside", "clear outside selection");
    flipXButton = getNewButton("FlipX", "flip selection about X axis");
    flipYButton = getNewButton("FlipY", "flip selection about Y axis");
    rotateClockwiseButton = getNewButton("Rotate+90", "rotate selection clockwise");
    rotateAntiClockwiseButton = getNewButton("Rotate-90", "rotate selection anti clockwise");
    selectAllButton = getNewButton("SelectAll", "select all live cells");
    centerButton = getNewButton("Center", "center selection");
    saveButton = getNewButton("SaveSelection", "write selection to file");
    selectionRandomButton = getNewButton("RandomSelection", "set selection to random states");
    moveCopyButton = getNewButton("Cut", "set selector to copy mode");
  }

  private JButton getNewButton(String name, String hint)
  {
    JButton b = new JButton(new ImageIcon(SquareCell.class.getResource("Resources/" + name + ".png")));
    b.setBackground(new Color(192, 192, 192));
    b.setToolTipText(hint);
    b.addActionListener(this);
    toolBar.add(b);
    return b;
  }

  private void centerDialog(JDialog dialog)
  {
    Point p = mainFrame.getLocation();
    int x = (int)p.getX() + (mainFrame.getWidth() - dialog.getWidth()) / 2;
    int y = (int)p.getY() + (mainFrame.getHeight() - dialog.getHeight()) / 2;
    dialog.setLocation(x, y);
  }

  public void setNewWorld()
  {
    newWorld = new NewWorld(mainFrame, rows, columns);
    centerDialog(newWorld);
    newWorld.setVisible(true);
    if (newWorld.hasChanged())
    {
      rows = newWorld.getRows();
      columns = newWorld.getColumns();
      width = mainFrame.getWidth();
      height = mainFrame.getHeight();
      makeNewSquareCell();
    }
  }

  private static void makeNewSquareCell()
  {
    mainFrame.setVisible(false);
    mainFrame.dispose();
    main(null);
  }

  public ColorSelector getColorSelector()
  {
    return colorSelector;
  }

  public StatusBar getStatusBar()
  {
    return statusBar;
  }

  public void setCurrentState(int state)
  {
    currentState = state;
  }

  public int getCurrentState()
  {
    return currentState;
  }

  public int getNoStates()
  {
    return rule.getNoStates();
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    display.makeGrid();
    statusBar.setDisplaySize(display.getRows(), display.getColumns());
    int rMax = rows - display.getRows();
    if (rowOffset > rMax)
      rowOffset = rMax;
    int cMax = columns - display.getColumns();
    if (columnOffset > cMax)
      columnOffset = cMax;
    drawDisplay();
  }

  public int getNeighbor(int r, int c)
  {
    return cellsCopy[(r + rows) % rows][(c + columns) % columns];
  }

  public void step()
  {
    byte t[][] = cellsCopy;
    cellsCopy = cells;
    cells = t;
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        cells[r][c] = (byte)rule.step(r, c);
    statusBar.setCycle(++cycle);
  }

  public byte[][] analyzerStep()
  {
    step();
    return cells;
  }

  private void clear()
  {
    cycle = 0;
    statusBar.setCycle(cycle);
    rowOffset = 0;
    columnOffset = 0;
    statusBar.setDisplayOffset(rowOffset, columnOffset);
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        cells[r][c] = 0;
    drawDisplay();
  }

  private boolean isClear()
  {
    boolean clear = true;
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        if (cells[r][c] > 0)
          clear = false;
    return clear;
  }

  private void setRandomParameters()
  {
    if (randomAll)
    {
      randomMinState = 0;
      randomMaxState = rule.getNoStates() - 1;
    }
    else
    {
      int noStates = rule.getNoStates();
      if (randomMaxState > noStates)
        randomMaxState = noStates;
      if (randomMinState > randomMaxState)
        randomMinState = randomMaxState;
    }
  }

  public void changeRandomOptions(JFrame mainFrame)
  {
    setRandomParameters();
    RandomOptionsDialog randomOptionsDialog = new RandomOptionsDialog(mainFrame, this);
    randomOptionsDialog.setData(this);
    centerDialog(randomOptionsDialog);
    randomOptionsDialog.setVisible(true);
    if (randomOptionsDialog.hasChanged())
      randomOptionsDialog.getData(this);
  }

  public int getRandomDensity()
  {
    return randomDensity;
  }

  public boolean getRandomClear()
  {
    return randomClear;
  }

  public boolean getRandomAll()
  {
    return randomAll;
  }

  public int getRandomMinState()
  {
    return randomMinState;
  }

  public int getRandomMaxState()
  {
    return randomMaxState;
  }

  public void setRandomDensity(int density)
  {
    randomDensity = density;
  }

  public void setRandomClear(boolean value)
  {
    randomClear = value;
  }

  public void setRandomAll(boolean value)
  {
    randomAll = value;
  }

  public void setRandomMinState(int value)
  {
    randomMinState = value;
  }

  public void setRandomMaxState(int value)
  {
    randomMaxState = value;
  }

  public void random(int rows, int columns, byte cells[][])
  {
    Random generator = new Random();
    if (randomClear)
      for (int r = 0; r < rows; r++)
        for (int c = 0; c < columns; c++)
          cells[r][c] = 0;
    setRandomParameters();
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        if (generator.nextInt(100) < randomDensity)
          cells[r][c] = (byte)(generator.nextInt(randomMaxState - randomMinState + 1) + randomMinState);
    drawDisplay();
  }

  private void selectAll()
  {
    if (selector.getState() == Selector.State.GroupSelected)
      selector.restore(rows, columns, rowOffset, columnOffset, cells);
    int r1 = rows;
    int c1 = columns;
    int rn = 0;
    int cn = 0;
    boolean clear = true;
    for (int r = 0; r < rows; r++)
      for (int c = 0; c < columns; c++)
        if (cells[r][c] > 0)
        {
          clear = false;
          if (r < r1)
            r1 = r;
          if (r > rn)
            rn = r;
          if (c < c1)
            c1 = c;
          if (c > cn)
            cn = c;
        }
    if (! clear)
    {
      selector.setTopLeft(r1 - rowOffset, c1 - columnOffset);
      selector.save(rn - rowOffset, cn - columnOffset, rowOffset, columnOffset, cells);
      selector.setState(Selector.State.GroupSelected);
      display.setSelected(true);
      drawDisplay();
    }
  }

  private void center()
  {
    if ((selector.getState() != Selector.State.GroupSelected) ||
       (selector.getRows() > display.getRows()) ||
       (selector.getColumns() > display.getColumns()))
      return;
    int r = (display.getRows() - selector.getRows()) / 2;
    int c = (display.getColumns() - selector.getColumns()) / 2;
    selector.move(r, c);
    drawDisplay();
  }

  private void analyzePattern()
  {
    analyzer = new Analyzer(mainFrame, this);
    if (analyzer.save(cells, rows, columns))
      return;
    analyzerButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Stop.png")));
    analyzerButton.setToolTipText("stop the analyzer");
    cycle = 0;
    runThread = new Thread(analyzer);
    analyzing = true;
    runThread.start();
  }

  public void showAnalysisResults()
  {
    analyzing = false;
    analyzerButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Analyzer.png")));
    analyzerButton.setToolTipText("show ship, oscillator or still life characteristics");
    centerDialog(analyzer);
    analyzer.setVisible(true);
    drawDisplay();
  }

  public boolean isAnalyzing()
  {
    return analyzing;
  }

  private void changeColors()
  {
    centerDialog(colorSelector);
    colorSelector.setVisible(true);
  }

  private void readPattern()
  {
    JFileChooser fc = new JFileChooser();
    fc.setCurrentDirectory(currentDirectory);
    if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
    {
      File f = fc.getSelectedFile();
      currentDirectory = fc.getCurrentDirectory();
      patterns = new Patterns(currentDirectory, f.getName());
      readFile(f);
    }
  }

  private void readFile(File f)
  {
    try
    {
      BufferedReader br = new BufferedReader(new FileReader(f));
      rule.readRule(br);
      int r = new Integer(br.readLine().substring(8).trim());
      int c = new Integer(br.readLine().substring(11).trim());
      if ((r > rows) || (c > columns))
        return;
      selector.setTopLeft(0, 0);
      selector.makeNewSelector(r - 1, c - 1, rowOffset, columnOffset);
      selector.setState(Selector.State.GroupSelected);
      display.setSelected(true);
      selector.readSelection(br);
      br.close();
      center();
      drawDisplay();
      statusBar.setPatternName(f.getName());
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }

  private void writePattern()
  {
    if (selector.getState() != Selector.State.GroupSelected)
      return;
    JFileChooser fc = new JFileChooser();
    fc.setCurrentDirectory(currentDirectory);
    if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION)
    {
      try
      {
        File f = fc.getSelectedFile();
        currentDirectory = fc.getCurrentDirectory();
        PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(f)));
        rule.writeRule(pw);
        pw.println("#Rows = " + selector.getRows());
        pw.println("#Columns = " + selector.getColumns());
        selector.writeSelection(pw);
        pw.close();
        patterns = new Patterns(currentDirectory, f.getName());
        statusBar.setPatternName(f.getName());
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }
  }

  public void drawDisplay()
  {
    if (! running)
    {
      int p = 0;
      for (int r = 0; r < rows; r++)
        for (int c = 0; c < columns; c++)
          if (cells[r][c] > 0)
            p++;
      statusBar.setPopulation(p);
    }
    for (int r = 0; r < display.getRows(); r++)     
      for (int c = 0; c < display.getColumns(); c++)
        display.setState(r, c, cells[r + rowOffset][c + columnOffset]);
    display.repaint();
  }
  
  private void setCut(boolean cut)
  {
    if (cut)
    {
      moveCopyButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Cut.png")));
      moveCopyButton.setToolTipText("set selector to copy mode");
      selector.setCut(true);
    }
    else
    {
      moveCopyButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Copy.png")));
      moveCopyButton.setToolTipText("set selector to move mode");
      selector.setCut(false);
    }
  }

  private void waitForThreadToEnd()
  {
    while (runThread.isAlive())
    {
      try
      {
        Thread.sleep(50);
      }
      catch (InterruptedException ie)
      {
      }
    }
  }

  private void startRunner()
  {
    runStopButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Stop.png")));
    runStopButton.setToolTipText("stop the simulation");
    cycle = 0;
    statusBar.setCycle(cycle);
    statusBar.setSpeedDelay(delay);
    statusBar.removePopulation();
    Runnable r = new CaRunner(this);
    runThread = new Thread(r);
    running = true;
    runThread.start();
  }

  private void stopRunner()
  {
    running = false;
    waitForThreadToEnd();
    statusBar.removeSpeedDelay();
    runStopButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Run.png")));
    runStopButton.setToolTipText("run the simulation");
    drawDisplay();
  }

  public boolean isRunning()
  {
    return running;
  }

  public int getDelay()
  {
    return delay;
  }

  public void actionPerformed(ActionEvent e)
  {
    Object source = e.getSource();
    if (source == quitButton)
    {
      System.exit(0);
    }
    if (analyzing)
    {
      if (source == analyzerButton)
      {
        analyzing = false;
        waitForThreadToEnd();
        analyzerButton.setIcon(new ImageIcon(SquareCell.class.getResource("Resources/Analyzer.png")));
        analyzerButton.setToolTipText("show ship, oscillator or still life characteristics");
      }
    }
    else if (running)
    {
      if ((source == runStopButton) || (source == newRuleButton) || (source == randomButton))
      {
        stopRunner();
      }
      if (source == newRuleButton)
      {
        rule.setNewRule();
        random(rows, columns, cells);
        startRunner();
      }
      else if (source == randomButton)
      {
        random(rows, columns, cells);
        startRunner();
      }
    }
    else if (source == runStopButton)
    {
      startRunner();
    }
    else if (source == newRuleButton)
    {
      rule.setNewRule();
      random(rows, columns, cells);
      startRunner();
    }
    else if (source == setNewWorldButton)
    {
      setNewWorld();
    }
    else if (source == openFileButton)
    {
      readPattern();
    }
    else if (source == saveFileButton)
    {
      if (! isClear())
      {
        selectAll();
        writePattern();
        selector.setState(Selector.State.None);
        display.setSelected(false);
        selector.restore(rows, columns, rowOffset, columnOffset, cells);
        drawDisplay();
      }
    }
    else if (source == stepButton)
    {
      step();
      drawDisplay();
    }
    else if (source == randomButton)
    {
      random(rows, columns, cells);
    }
    else if (source == randomOptionsButton)
    {
      changeRandomOptions(mainFrame);
    }
    else if (source == clearButton)
    {
      selector.resetSelector();
      display.setSelected(false);
      statusBar.removePatternName();
      clear();
    }
    else if (source == analyzerButton)
    {
      analyzePattern();
    }
    else if (source == changeColorsButton)
    {
      changeColors();
    }
    else if (source == rulesButton)
    {
      rule.changeRule();
    }
    else if (source == ruleOptionsButton)
    {
      rule.setOptions();
    }
    else if (source == clearInsideButton)
    {
      selector.clear();
      selector.resetSelector();
      display.setSelected(false);
      drawDisplay();
    }
    else if (source == clearOutsideButton)
    {
      if (selector.getState() == Selector.State.GroupSelected)
        clear();
    }
    else if (source == flipXButton)
    {
      selector.flipX();
      drawDisplay();
    }
    else if (source == flipYButton)
    {
      selector.flipY();
      drawDisplay();
    }
    else if (source == rotateClockwiseButton)
    {
      selector.rotateClockwise();
      drawDisplay();
    }
    else if (source == rotateAntiClockwiseButton)
    {
      selector.rotateAntiClockwise();
      drawDisplay();
    }
    else if (source == selectAllButton)
    {
      selectAll();
      setCut(true);
    }
    else if (source == centerButton)
    {
      center();
    }
    else if (source == saveButton)
    {
      writePattern();
    }
    else if (source == selectionRandomButton)
    {
      selector.random();
      drawDisplay();
    }
    else if (source == moveCopyButton)
    {
      if (selector.getCut())
      {
        setCut(false);
      }
      else
      {
        setCut(true);
      }
    }
    requestFocus();
  }

  public void mousePressed(MouseEvent e)
  {
    if (running)
      return;
    int r = display.yCoordToRowNumber(e.getY());
    int c = display.xCoordToColumnNumber(e.getX());
    if ((r < 0) || (c < 0))
      return;
    if (e.isControlDown())
    {
      if (selector.getState() == Selector.State.None)
      {
        selector.setState(Selector.State.TopLeftMarked);
        display.setSelected(true);
        selector.setTopLeft(r, c);
        drawDisplay();
      }
      else if (selector.getState() == Selector.State.TopLeftMarked)
      {
        selector.setState(Selector.State.GroupSelected);
        selector.save(r, c, rowOffset, columnOffset, cells);
        setCut(true);
        drawDisplay();
      }
      else if (selector.getState() == Selector.State.GroupSelected)
      {
        selector.setState(Selector.State.None);
        display.setSelected(false);
        selector.restore(rows, columns, rowOffset, columnOffset, cells);
        drawDisplay();
      }
    }
    else if (selector.getState() == Selector.State.None)
    {
      if (e.isShiftDown())
      {
        currentState = (currentState + 1) % rule.getNoStates();
        colorSelector.showCurrentState();
      }
      display.setState(r, c, currentState);
      cells[r + rowOffset][c + columnOffset] = (byte)currentState;
      statusBar.setCurrentState(display.getState(r, c));
      drawDisplay();
    }
  }

  public void mouseDragged(MouseEvent e)
  {
    if (running || (selector.getState() != Selector.State.GroupSelected)
        || e.isShiftDown() || e.isControlDown())
      return;
    int r = display.yCoordToRowNumber(e.getY());
    int c = display.xCoordToColumnNumber(e.getX());
    if ((r < 0) || (c < 0))
      return;
    int rMax = rows - selector.getRows() - rowOffset;
    if (r > rMax)
      r = rMax;
    int cMax = columns - selector.getColumns() - columnOffset;
    if (c > cMax)
      c = cMax;
    selector.move(r, c);
    statusBar.removeCurrentState();
    statusBar.setCursorLocation(r + rowOffset, c + columnOffset);
    drawDisplay();
  }

  public void mouseMoved(MouseEvent e)
  {
    int r = display.yCoordToRowNumber(e.getY());
    int c = display.xCoordToColumnNumber(e.getX());
    if ((r < 0) || (c < 0))
    {
      statusBar.removeCursorLocation();
      statusBar.removeCurrentState();
    }
    else
    {
      statusBar.setCursorLocation(r + rowOffset, c + columnOffset);
      statusBar.setCurrentState(display.getState(r, c));
    }
  }

  public void mouseExited(MouseEvent e)
  {
    statusBar.removeCursorLocation();
    statusBar.removeCurrentState();
  }

  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}

  public void keyPressed(KeyEvent e)
  {
    int keyCode = e.getKeyCode();
    if (running)
    {
      if (keyCode == KeyEvent.VK_ENTER)
      {
        stopRunner();
        return;
      }
      else if (keyCode == KeyEvent.VK_N)
      {
        stopRunner();
        rule.setNewRule();
        random(rows, columns, cells);
        startRunner();
      }
      else if (keyCode == KeyEvent.VK_R)
      {
        stopRunner();
        random(rows, columns, cells);
        startRunner();
      }
      else if (keyCode == KeyEvent.VK_ADD)
      {
        if (delay >= 10)
          delay -= 10;
        statusBar.setSpeedDelay(delay);
      }
      else if (keyCode == KeyEvent.VK_SUBTRACT)
      {
        delay += 10;
        statusBar.setSpeedDelay(delay);
      }
      return;
    }
    int increment = (e.isShiftDown()) ? 25 : 1;
    if (keyCode == KeyEvent.VK_ENTER)
    {
      if (! analyzing)
        startRunner();
    }
    else if (keyCode == KeyEvent.VK_SPACE)
    {
      step();
      drawDisplay();
    }
    else if (keyCode == KeyEvent.VK_N)
    {
      rule.setNewRule();
      random(rows, columns, cells);
      startRunner();
    }
    else if (keyCode == KeyEvent.VK_R)
    {
      random(rows, columns, cells);
    }
    else if (keyCode == KeyEvent.VK_LEFT)
    {
      if (columnOffset > 0)
      {
        if (columnOffset < increment)
          increment = columnOffset;
        columnOffset -= increment;
        if (selector.getState() != Selector.State.None)
          selector.moveRight(increment);
        drawDisplay();
      }
    }
    else if (keyCode == KeyEvent.VK_RIGHT)
    {
      int cMax = columns - display.getColumns();
      if (columnOffset < cMax)
      {
        if ((columnOffset + increment) > cMax)
          increment = cMax - columnOffset;
        columnOffset += increment;
        if (selector.getState() != Selector.State.None)
          selector.moveLeft(increment);
        drawDisplay();
      }
    }
    else if (keyCode == KeyEvent.VK_UP)
    {
      if (rowOffset > 0)
      {
        if (rowOffset < increment)
          increment = rowOffset;
        rowOffset -= increment;
        if (selector.getState() != Selector.State.None)
          selector.moveDown(increment);
        drawDisplay();
      }
    }
    else if (keyCode == KeyEvent.VK_DOWN)
    {
      int rMax = rows - display.getRows();
      if (rowOffset < rMax)
      {
        if ((rowOffset + increment) > rMax)
          increment = rMax - rowOffset;
        rowOffset += increment;
        if (selector.getState() != Selector.State.None)
          selector.moveUp(increment);
        drawDisplay();
      }
    }
    else if (keyCode == KeyEvent.VK_ADD)
    {
      cellSize++;
      display.setSquareSize(cellSize);
      drawDisplay();
    }
    else if (keyCode == KeyEvent.VK_SUBTRACT)
    {
      if (cellSize > 2)
        cellSize--;
      display.setSquareSize(cellSize);
      int rMax = rows - display.getRows();
      if (rowOffset > rMax)
        rowOffset = rMax;
      int cMax = columns - display.getColumns();
      if (columnOffset > cMax)
        columnOffset = cMax;
      drawDisplay();
    }
    else if (keyCode == KeyEvent.VK_PAGE_UP)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getPrevious();
        readFile(f);
      }
    }
    else if (keyCode == KeyEvent.VK_PAGE_DOWN)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getNext();
        readFile(f);
      }
    }
    else if (keyCode == KeyEvent.VK_HOME)
    {
      if ((patterns != null) && (patterns.getNumberPatterns() > 0))
      {
        clear();
        File f = patterns.getCurrent();
        readFile(f);
      }
    }
    statusBar.setDisplayOffset(rowOffset, columnOffset);
    statusBar.setDisplaySize(display.getRows(), display.getColumns());
  }

  public void keyReleased(KeyEvent e) {}
  public void keyTyped(KeyEvent e) {}
}
